from typing import Optional
import logging
from io import BytesIO
import pandas as pd
import views_schema
from . import remotes

logger = logging.getLogger(__name__)

def check_parquet_response(response):
    if response.headers.get("Content-Type") == "application/octet-stream":
        pass
    else:
        raise remotes.RemoteError(
                f"Remote {response.url} returned wrong content type: "
                f"{response.headers.get('Content-Type')}"
                )

def fetch_queryset(
        remotes_api: remotes.Api,
        name,
        start_date:Optional[str]=None,
        end_date:Optional[str]=None):

    response = remotes_api.http(
            "GET",
            ("data",name),
            {"start_date":start_date,"end_date": end_date}
            )

    check_parquet_response(response)
    return pd.read_parquet(BytesIO(response.content))

def post_queryset(remotes_api: remotes.Api, queryset: views_schema.Queryset):
    """
    Post a queryset to the remote API
    """
    return remotes_api.http("POST",("queryset",),{},data=queryset.json()).json()

def put_queryset(remotes_api: remotes.Api, name:str,queryset: views_schema.Queryset):
    """
    Put a queryset to the remote API, overwriting the existing queryset
    """
    return remotes_api.http("PUT",("queryset",name),{},data=queryset.json()).json()

update_queryset = lambda remotes_api, queryset: put_queryset(remotes_api, queryset.name, queryset)

def list_querysets(remotes_api: remotes.Api):
    """
    List available querysets
    """
    return remotes_api.http("GET",("queryset",),{}).json()

def delete_queryset(remotes_api: remotes.Api, name):
    """
    Delete a named queryset
    """
    return remotes_api.http("DELETE",("queryset",name),{})

def show_queryset(remotes_api: remotes.Api, name:str):
    """
    Show details about a queryset
    """
    return remotes_api.http("GET",("queryset",name),{}).json()



"""
High-level operations intended for users
"""
from importlib.metadata import version
from datetime import date
from typing import Optional
import logging
import time
from functools import wraps
from toolz.functoolz import curry, compose
from requests import HTTPError
import views_schema
from . import settings,crud,remotes,checks

logger = logging.getLogger(__name__)

remotes_api = remotes.Api(settings.config_get("REMOTE_URL"),{})


get_latest_version = curry(remotes.latest_pyproject_version, settings.config_get("REPO_URL"))
check_remote_version = curry(checks.check_remote_version, remotes_api)

def check_latest_version():
    latest_known = settings.config_get("LATEST_KNOWN_VERSION")
    latest_actual = get_latest_version()
    current = version("viewser")

    if latest_known < latest_actual:
        settings.config_set_in_file("LATEST_KNOWN_VERSION",latest_actual)

    if current < latest_actual:
        logger.warning(
                f"There is a newer version of viewser available ({latest_actual}), "
                "download with pip install --upgrade viewser. "
                f"Current version is {current}."
                )
    else:
        logger.debug("Up to date!")


def check_remotes(fn):
    """
    Wrapper used for version checking when calling remote APIs
    """

    @wraps(fn)
    def inner(*args,**kwargs):
        check_latest_version()
        check_remote_version()
        return fn(*args,**kwargs)
    return inner

def publish(queryset: views_schema.Queryset):
    """
    Publishes a queryset to the ViEWSER databases, which can then be fetched
    by calling `viewser.operations.fetch(queryset.name)`
    """

    try:
        crud.post_queryset(remotes_api, queryset)
    except HTTPError as httpe:
        logger.info("Queryset named \"%s\" exists, updating",
                queryset.name
            )
    else:
        return

    try:
        crud.update_queryset(remotes_api, queryset)
    except HTTPError as httpe:
        logger.critical("Update returned %s: %s",
                str(httpe.response.status_code),
                str(httpe.response.content),
            )

def fetch(queryset_name:str, start_date: Optional[date] = None, end_date: Optional[date] = None):
    """
    Fetches data for a queryset,
    """
    retries = 0
    while retries < settings.config_get("RETRIES"):
        try:
            return crud.fetch_queryset(remotes_api, queryset_name, start_date, end_date)
        except remotes.OperationPending:
            logger.info("Queryset \"%s\" is being compiled... (%s retries)",
                    queryset_name, str(retries)
                    )
            retries += 1
            time.sleep(settings.config_get("RETRY_FREQUENCY"))


api_call = lambda fn: check_remotes(curry(fn, remotes_api))

post_queryset = api_call(crud.post_queryset)
put_queryset = api_call(crud.put_queryset)
list_querysets = api_call(crud.list_querysets)
delete_queryset = api_call(crud.delete_queryset)
show_queryset = api_call(crud.show_queryset)



import os
import functools
import logging
import json
import fitin

from toolz.functoolz import compose,curry

logger = logging.getLogger(__name__)

REQUIRED_SETTINGS = (
            "REMOTE_URL",
        )

DEFAULT_SETTINGS = {
        "RETRY_FREQUENCY": 5, # seconds
        "RETRIES": 80,

        "HANDSHAKE_PATH": "",
        "REPO_URL": "https://www.github.com/prio-data/viewser",
        "LATEST_KNOWN_VERSION": "0.0.0"
    }

CONFIG_DIR = os.path.expanduser("~/.views")
SETTINGS_FILE_PATH = os.path.join(CONFIG_DIR, "config.json")

def log_decorator(msg,level = "debug"):
    def wrapper(fn):
        functools.wraps(fn)
        def inner(*args,**kwargs):
            getattr(logger,level)(msg)
            return fn(*args,**kwargs)
        return inner
    return wrapper




def load_config_from_file():
    with open(SETTINGS_FILE_PATH) as f:
        return json.load(f)

def save_config_to_file(cfg):
    with open(SETTINGS_FILE_PATH,"w") as f:
        json.dump(cfg,f)

def with_key_value(key,value,cfg):
    cfg[key] = value
    return cfg

config_set = lambda load,save,key,value: compose(
        save,
        curry(with_key_value,key,value),
        load,
        )()

config_set_in_file = curry(
        config_set,
        load_config_from_file,
        log_decorator("Altering config file")(save_config_to_file)
        )

def reset_defaults(load,save):
    with_defaults = [curry(with_key_value,k,v) for k,v in DEFAULT_SETTINGS.items()]
    compose(
            save,
            compose(*with_defaults),
            load,
        )()

reset_config_file_defaults = curry(
        reset_defaults,
        load_config_from_file,
        save_config_to_file
        )

try:
    os.makedirs(CONFIG_DIR)
except FileExistsError:
    pass

try:
    with open(SETTINGS_FILE_PATH) as settings_file:
        config_dict = json.load(settings_file)
except FileNotFoundError:
    logger.critical("Could not find config file, please run: 'viewser config interactive'")
    config_dict = {}
    save_config_to_file(config_dict)

config_get = fitin.seek_config([
    fitin.environs_resolver(),
    fitin.dict_resolver(config_dict),
    fitin.dict_resolver(DEFAULT_SETTINGS)
    ])

try:
    logging.basicConfig(level=getattr(logging,config_get("LOG_LEVEL")))
except(AttributeError, KeyError):
    logging.basicConfig(level=logging.INFO)

def configure_interactively():
    for key in REQUIRED_SETTINGS:
        config_set_in_file(
                key,
                input(f"{key} >> "),
            )




import json
from importlib.metadata import version
from unittest import TestCase
import httpretty
from viewser import checks,remotes

class TestChecks(TestCase):
    @httpretty.activate()
    def test_check_remote_version(self):

        current_version = version("viewser")
        maj, min, pat = (int(v) for v in current_version.split("."))

        httpretty.register_uri(httpretty.GET, "http://test/", body = json.dumps({
                "viewser_version": f"{maj-1}.{min}.{pat}"
            }))

        self.assertRaises(
                checks.WrongVersion,
                checks.check_remote_version(remotes.Api("http://test",{}))(lambda: None)
            )

        httpretty.register_uri(httpretty.GET, "http://test/", body = json.dumps({
                "viewser_version": f"{maj+1}.{min}.{pat}"
            }))

        self.assertRaises(
                checks.WrongVersion,
                checks.check_remote_version(remotes.Api("http://test",{}))(lambda: None)
            )

        httpretty.register_uri(httpretty.GET, "http://test/", body = json.dumps({
                "viewser_version": f"{maj}.{min}.{pat}"
            }))

        checks.check_remote_version(remotes.Api("http://test",{}))(lambda: None)()




from importlib.metadata import version
import logging
import json
from datetime import datetime
from typing import Optional
import io
import click
import tabulate
import requests
import views_schema
from . import settings, operations, remotes

logger = logging.getLogger(__name__)

@click.group()
@click.option("--debug/--no-debug", default=False, help="Display debug logging messages")
def viewser(debug: bool):
    if debug:
        logging.getLogger().setLevel(logging.DEBUG)

@viewser.group(name="queryset", short_help="operations related to querysets")
def queryset():
    pass

@queryset.command(name="fetch", short_help="fetch data for a queryset")
@click.argument("name")
@click.argument("out-file", type=click.File("wb"))
@click.option("-s","--start-date", type=click.DateTime())
@click.option("-e","--end-date", type=click.DateTime())
def queryset_fetch(
        name:str,
        out_file:io.BufferedWriter,
        start_date: Optional[datetime],
        end_date: Optional[datetime]):
    """
    Fetch data for a queryset named NAME from ViEWS cloud and save it to OUT_FILE
    """
    click.echo("Fetching queryset {name}...")
    operations.fetch(name,start_date,end_date).to_parquet(out_file)
    click.echo("Saved to {out_file}")

@queryset.command(name="list", short_help="show a list of available querysets")
@click.option("--as-json/--as-table", default=False, help="output results as json")
def queryset_list(as_json: bool):
    """
    Show a list of available querysets.
    """

    querysets = operations.list_querysets()
    if as_json:
        click.echo(querysets)
    else:
        click.echo(tabulate.tabulate([[qs] for qs in querysets["querysets"]],headers=["name"]))

@queryset.command(name="show", short_help="show details about a queryset")
@click.argument("name", type=str)
def queryset_show(name: str):
    """
    Show detailed information about a queryset
    """
    qs = operations.show_queryset(name)
    click.echo(json.dumps(qs, indent=4))

@queryset.command(name="delete", short_help="delete a queryset")
@click.confirmation_option(prompt="Delete queryset?")
@click.argument("name", type=str)
def queryset_delete(name: str):
    """
    Delete a queryset.
    """
    try:
        operations.delete_queryset(name)
        click.echo(f"Deleted {name}")
    except requests.HTTPError as httpe:
        raise click.ClickException(
                "Delete operation returned "
                f"{httpe.response.status_code}: "
                f"'{httpe.response.content.decode()}'"
            )


@queryset.command(name="upload", short_help="upload a queryset")
@click.argument("queryset_file", type=click.File("r","utf-8"))
@click.option("-n", "--name", type=str)
@click.option("--overwrite/--no-overwrite",default = False)
def queryset_upload(
        queryset_file: io.BufferedReader,
        name: Optional[str],
        overwrite: bool):
    """
    Upload a queryset defined in a JSON file. The name of the queryset can be
    overridden with the --name flag.
    """
    if name is not None:
        click.echo(f"Renaming to {name}")

    if overwrite:
        click.echo("Overwriting!")

    qs = views_schema.Queryset(**json.load(queryset_file))
    try:
        click.echo(operations.post_queryset(qs))
    except requests.HTTPError as httpe:
        click.echo(httpe.response.content)

@viewser.group(name="config", short_help="configure viewser")
def config():
    """
    Configure viewser
    """

@config.command(name="interactive", short_help="interactively configure viewser")
def config_interactive():
    """
    Interactively configure viewser (useful for first-time config)
    """
    settings.configure_interactively()
    click.echo("All done!")

@config.command(name="set", short_help="set a configuration value")
@click.argument("name", type=str)
@click.argument("value", type=str)
@click.option("--override/--no-override",default=True)
def config_set(name: str, value: str, override: bool): # pylint: disable=redefined-builtin
    """
    Set a configuration value.
    """

    overrides = True
    try:
        settings.config_get(name)
    except KeyError:
        overrides = False

    if not override and overrides:
        click.echo(f"Setting {name} already set, override not specified (see --help)")
        return

    settings.config_set_in_file(name,value)
    click.echo(f"{name}: {value}")

@config.command(name="get", short_help="get a configuration value")
@click.argument("name", type=str)
def config_get(name: str):
    """
    Get a configuration value.
    """
    click.echo(f"{name}: {settings.config_get(name)}")

@config.command(name="reset", short_help="reset default config values")
@click.confirmation_option(prompt="Reset config?")
def config_reset():
    """
    Reset config, writing default values over current values.
    """
    settings.reset_config_file_defaults()
    click.echo("Config file reset")

@config.command(name="list", short_help="show all configuration settings")
def config_list():
    """
    Show all current configuration values
    """
    click.echo(tabulate.tabulate(settings.config_dict.items()))

@viewser.group(name="help", short_help="commands for finding documentation")
def get_help():
    """
    Commands related to finding documentation.
    """

@get_help.command(name = "wiki")
def wiki():
    """
    Open the viewser wiki in a new browser window, which contains all you need
    to know about using viewser.
    """
    remotes.browser(settings.config_get("REPO_URL"),"wiki")

@get_help.command(name = "issue")
def issue():
    """
    Opens the viewser "new issue" page in a new browser window.
    """
    remotes.browser(
            settings.config_get("REPO_URL"),
            "issue","new",
            body=f"My viewser version is {version('viewser')}"
        )




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['viewser', 'viewser.tests']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.1,<9.0.0',
 'crayons>=0.4.0,<0.5.0',
 'environs>=9.3.1,<10.0.0',
 'fitin>=0.2.0,<0.3.0',
 'pandas>=1.2.3,<2.0.0',
 'pyarrow>=3.0.0,<4.0.0',
 'pydantic>=1.8.1,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'tabulate>=0.8.9,<0.9.0',
 'toml>=0.10.2,<0.11.0',
 'toolz>=0.11.1,<0.12.0',
 'views-schema>=0.1.0,<0.2.0']

entry_points = \
{'console_scripts': ['viewser = viewser.cli:viewser']}

setup_kwargs = {
    'name': 'viewser',
    'version': '2.0.2',
    'description': '',
    'long_description': None,
    'author': 'peder2911',
    'author_email': 'pglandsverk@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)




import os
import webbrowser
import logging
from urllib import parse
import toml
import requests

logger = logging.getLogger(__name__)

class OperationPending(Exception):
    pass

class RemoteError(Exception):
    pass

def check_response(response):
    if response.status_code == 200:
        pass
    elif response.status_code == 202:
        raise OperationPending
    else:
        raise requests.HTTPError(response=response)

class Api:
    def __init__(self,url,paths=None):
        self._base_url = url
        self.paths = paths

    def url(self,*args,**kwargs):
        url = os.path.join(self._base_url,*args)

        kwargs = {k:v for k,v in kwargs.items() if v is not None}
        if kwargs:
            url += "?" + parse.urlencode(kwargs)
        return url

    @staticmethod
    def check_response(response):
        if response.status_code == 202:
            raise OperationPending

        if str(response.status_code)[0] == "2":
            pass
        elif str(response.status_code)[0] == "5":
            raise RemoteError(
                    f"{response.url} returned {response.status_code}: "
                    f"{response.content.decode()}"
                    )
        else:
            raise requests.HTTPError(response=response)

    def http(self,method,path,parameters,*args,**kwargs):
        url = self.url(*path,**parameters)
        logger.debug("Requesting url %s",url)
        rsp = requests.request(method,url,*args,**kwargs)
        self.check_response(rsp)
        return rsp

def browser(base,*args,**kwargs):
    webbrowser.open(Api(base).url(*args,**kwargs))

def latest_pyproject_version(repo_url):
    """
    Gets the latest version of the CLI from Github
    """
    repo_path = parse.urlparse(repo_url).path
    url = parse.urljoin("https://raw.githubusercontent.com/",repo_path)
    url = os.path.join(url,"master/pyproject.toml")

    rsp = requests.get(url)
    Api.check_response(rsp)
    pyproject = toml.loads(rsp.content.decode())
    version = pyproject["tool"]["poetry"]["version"]
    logger.debug("Latest version from github: %s",version)
    return version





        
import functools
from importlib.metadata import version
import re
import json
import logging
from . import exceptions, remotes

logger = logging.getLogger(__name__)

class WrongVersion(Exception):
    """
    Raised when remotes call for a different major version than the one currently installed
    """

def check_remote_version(remotes_api: remotes.Api):
    """
    Used to decorate operations that connect to remote API(s).  The remotes
    should define a "handshake" path, which should return json containing a
    "viewser_version" field.
    """
    logger.debug("Checking remote version")
    rsp = remotes_api.http("GET", ("",), {})
    if rsp.status_code != 200:
        raise exceptions.ConfigurationError(
                f"The handshake endpoint returned {rsp.status_code}. "
                "Is the handshake URL {url} correct?"
                )
    try:
        remote_version = rsp.json()["viewser_version"]
    except (json.JSONDecodeError,KeyError) as e:
        raise exceptions.ConfigurationError(
                "The handshake endpoint did not return the right data. "
                f"Is the handshake URL correct? {remotes_api.url('')}"
                ) from e

    try:
        assert re.search(r"[0-9]+\.[0-9]+\.[0-9]+",remote_version)
    except AssertionError as ae:
        raise remotes.RemoteError(
                "Handshake version wrong format: {remote_version} (expected n.n.n)"
                ) from ae

    major_version = lambda x: x.split(".")[0]

    logger.debug(f"%s vs %s",remote_version, version("viewser"))
    if not major_version(remote_version) == major_version(version("viewser")):
        raise WrongVersion(
                "Viewser installation has wrong major version. "
                f"Local={version('viewser')} / Remote={remote_version}"
                )
